def null_padded(string, length):
    """
    Force a string to a given length by right-padding it with zero-bytes,
    clipping the initial string if neccessary.
    """
    return string[:length] + ('\0' * (length - len(string)))



from constants import END_OF_DATA_MARKER
from constants import END_OF_FILE_MARKER
from constants import END_OF_REPLAY_FILE_MARKER
from constants import TOP10_MULTIPLAYER
from constants import TOP10_SINGLEPLAYER
from models import Frame
from models import GroundTouchAEvent
from models import GroundTouchBEvent
from models import LeftVoltEvent
from models import Level
from models import Obj
from models import ObjectTouchEvent
from models import Picture
from models import Point
from models import Polygon
from models import Replay
from models import RightVoltEvent
from models import TurnEvent
from utils import null_padded
import random
import struct


packers = {
    'Point': lambda point:
        struct.pack('d', point.x) + struct.pack('d', point.y),
    'Polygon': lambda polygon: ''.join([
        struct.pack('I', polygon.grass),
        struct.pack('I', len(polygon.points)),
    ] + [pack_level(point) for point in polygon.points]),
    'Obj': lambda obj: ''.join([
        pack_level(obj.point),
        struct.pack('I', obj.type),
        struct.pack('I', obj.gravity),
        struct.pack('I', obj.animation_number)]),
    'Picture': lambda picture: ''.join([
        null_padded(picture.picture_name, 10),
        null_padded(picture.texture_name, 10),
        null_padded(picture.mask_name, 10),
        pack_level(picture.point),
        struct.pack('I', picture.distance),
        struct.pack('I', picture.clipping)])
}


def pack_level(item):
    """
    Pack a level-related item to its binary representation readable by
    Elastomania.
    """
    if type(item).__name__ in packers:
        return packers[type(item).__name__](item)

    level = item
    polygon_checksum = sum([sum([point.x + point.y
                                 for point in polygon.points])
                            for polygon in level.polygons])
    object_checksum = sum([obj.point.x + obj.point.y + obj.type
                           for obj in level.objects])
    picture_checksum = sum([picture.point.x + picture.point.y
                            for picture in level.pictures])
    collected_checksum = 3247.764325643 * (polygon_checksum +
                                           object_checksum +
                                           picture_checksum)
    integrity_1 = collected_checksum
    integrity_2 = random.randint(0, 5871) + 11877 - collected_checksum
    integrity_3 = random.randint(0, 5871) + 11877 - collected_checksum
    integrity_4 = random.randint(0, 6102) + 12112 - collected_checksum
    assert ((integrity_3 - integrity_2) <= 5871)

    return ''.join([
        'POT14',
        struct.pack('H', level.level_id & 0xFFFF),
        struct.pack('I', level.level_id),
        struct.pack('d', integrity_1),
        struct.pack('d', integrity_2),
        struct.pack('d', integrity_3),
        struct.pack('d', integrity_4),
        null_padded(level.name, 51),
        null_padded(level.lgr, 16),
        null_padded(level.ground_texture, 10),
        null_padded(level.sky_texture, 10),
        struct.pack('d', len(level.polygons) + 0.4643643),
    ] + map(pack_level, level.polygons) + [
        struct.pack('d', len(level.objects) + 0.4643643),
    ] + map(pack_level, level.objects) + [
        struct.pack('d', len(level.pictures) + 0.2345672),
    ] + map(pack_level, level.pictures) + [
        struct.pack('I', END_OF_DATA_MARKER),
    ] + map(chr, TOP10_SINGLEPLAYER) + [
    ] + map(chr, TOP10_MULTIPLAYER) + [
        struct.pack('I', END_OF_FILE_MARKER),
    ])


def unpack_level(data):
    """
    Unpack a level-related item from its binary representation readable by
    Elastomania.
    """

    data = iter(data)

    def munch(n):
        return ''.join([next(data) for _ in range(n)])

    level = Level()
    assert munch(5) == 'POT14'
    munch(2)
    level.level_id = struct.unpack('I', munch(4))[0]
    munch(8 * 4)
    level.name = munch(51).rstrip('\0')
    level.lgr = munch(16).rstrip('\0')
    level.ground_texture = munch(10).rstrip('\0')
    level.sky_texture = munch(10).rstrip('\0')

    number_of_polygons = int(struct.unpack('d', munch(8))[0])
    for _ in range(number_of_polygons):
        grass = struct.unpack('I', munch(4))[0]
        number_of_vertices = struct.unpack('I', munch(4))[0]
        points = []
        for __ in range(number_of_vertices):
            x = struct.unpack('d', munch(8))[0]
            y = struct.unpack('d', munch(8))[0]
            points.append(Point(x, y))
        level.polygons.append(Polygon(points, grass=grass))

    number_of_objects = int(struct.unpack('d', munch(8))[0])
    for _ in range(number_of_objects):
        x = struct.unpack('d', munch(8))[0]
        y = struct.unpack('d', munch(8))[0]
        type = struct.unpack('I', munch(4))[0]
        gravity = struct.unpack('I', munch(4))[0]
        animation_number = struct.unpack('I', munch(4))[0]
        level.objects.append(Obj(Point(x, y),
                                 type,
                                 gravity=gravity,
                                 animation_number=animation_number))

    number_of_pictures = int(struct.unpack('d', munch(8))[0])
    for _ in range(number_of_pictures):
        picture_name = munch(10)
        texture_name = munch(10)
        mask_name = munch(10)
        x = struct.unpack('d', munch(8))[0]
        y = struct.unpack('d', munch(8))[0]
        distance = struct.unpack('I', munch(4))[0]
        clipping = struct.unpack('I', munch(4))[0]
        level.pictures.append(Picture(Point(x, y),
                                      picture_name=picture_name,
                                      texture_name=texture_name,
                                      mask_name=mask_name,
                                      distance=distance,
                                      clipping=clipping))
    return level


def unpack_replay(data):
    """
    Unpack a replay-related item from its binary representation readable by
    Elastomania.
    """

    data = iter(data)

    def munch(n):
        return ''.join([next(data) for _ in range(n)])

    def read_int32():
        return struct.unpack('i', munch(4))[0]

    def read_uint32():
        return struct.unpack('I', munch(4))[0]

    def read_int16():
        return struct.unpack('h', munch(2))[0]

    def read_uint8():
        return struct.unpack('B', munch(1))[0]

    def read_float():
        return struct.unpack('f', munch(4))[0]

    def read_double():
        return struct.unpack('d', munch(8))[0]

    replay = Replay()

    number_of_replay_frames = read_int32()
    munch(4)
    replay.is_multi = bool(read_int32())
    replay.is_flagtag = bool(read_int32())
    replay.level_id = read_uint32()
    replay.level_name = munch(12).rstrip('\0')
    munch(4)

    frame_numbers = range(number_of_replay_frames)
    xs = [read_float() for _ in frame_numbers]
    ys = [read_float() for _ in frame_numbers]
    left_wheel_xs = [read_int16() for _ in frame_numbers]
    left_wheel_ys = [read_int16() for _ in frame_numbers]
    right_wheel_xs = [read_int16() for _ in frame_numbers]
    right_wheel_ys = [read_int16() for _ in frame_numbers]
    head_xs = [read_int16() for _ in frame_numbers]
    head_ys = [read_int16() for _ in frame_numbers]
    rotations = [read_int16() for _ in frame_numbers]
    left_wheel_rotations = [read_uint8() for _ in frame_numbers]
    right_wheel_rotations = [read_uint8() for _ in frame_numbers]
    gas_and_turn_states = [read_uint8() for _ in frame_numbers]
    sound_effect_volumes = [read_int16() for _ in frame_numbers]

    for (x,
         y,
         left_wheel_x,
         left_wheel_y,
         right_wheel_x,
         right_wheel_y,
         head_x,
         head_y,
         rotation,
         left_wheel_rotation,
         right_wheel_rotation,
         gas_and_turn_state,
         sound_effect_volume) in zip(xs,
                                     ys,
                                     left_wheel_xs,
                                     left_wheel_ys,
                                     right_wheel_xs,
                                     right_wheel_ys,
                                     head_xs,
                                     head_ys,
                                     rotations,
                                     left_wheel_rotations,
                                     right_wheel_rotations,
                                     gas_and_turn_states,
                                     sound_effect_volumes):

        frame = Frame()
        frame.position = Point(x, y)
        frame.left_wheel_position = Point(left_wheel_x, left_wheel_y)
        frame.right_wheel_position = Point(right_wheel_x, right_wheel_y)
        frame.head_position = Point(head_x, head_y)
        frame.rotation = rotation
        frame.left_wheel_rotation = left_wheel_rotation
        frame.right_wheel_rotation = right_wheel_rotation
        frame.is_gasing = bool(gas_and_turn_state & 0b1)
        frame.is_turned_right = bool(gas_and_turn_state & 0b10)
        # preserve remaining 6 bits of state
        frame._gas_and_turn_state = gas_and_turn_state
        frame.spring_sound_effect_volume = sound_effect_volume
        replay.frames.append(frame)

    number_of_replay_events = read_int32()
    for _ in range(number_of_replay_events):
        event_time = read_double()
        event_type_1 = read_int32()
        event_type_2 = read_int32()

        if event_type_2 == 0:
            event = ObjectTouchEvent()
            event.object_number = event_type_1
        elif event_type_1 == 393215 and event_type_2 == 1065185444:
            event = TurnEvent()
        elif event_type_1 == 524287 and event_type_2 == 1065185444:
            event = LeftVoltEvent()
        elif event_type_1 == 458751 and event_type_2 == 1065185444:
            event = RightVoltEvent()
        elif event_type_1 == 131071:
            event = GroundTouchAEvent()
            event.value = event_type_2
        elif event_type_1 == 327679:
            event = GroundTouchBEvent()
            event.value = event_type_2

        event.time = event_time
        replay.events.append(event)

    return replay


def pack_replay(item):
    """
    Pack a replay-related item to its binary representation readable by
    Elastomania.
    """

    if isinstance(item, ObjectTouchEvent):
        return (struct.pack('d', item.time) +
                struct.pack('I', item.object_number) +
                struct.pack('I', 0))

    if isinstance(item, TurnEvent):
        return (struct.pack('d', item.time) +
                struct.pack('I', 393215) +
                struct.pack('I', 1065185444))

    if isinstance(item, LeftVoltEvent):
        return (struct.pack('d', item.time) +
                struct.pack('I', 524287) +
                struct.pack('I', 1065185444))

    if isinstance(item, RightVoltEvent):
        return (struct.pack('d', item.time) +
                struct.pack('I', 458751) +
                struct.pack('I', 1065185444))

    if isinstance(item, GroundTouchAEvent):
        return (struct.pack('d', item.time) +
                struct.pack('I', 131071) +
                struct.pack('I', item.value))

    if isinstance(item, GroundTouchBEvent):
        return (struct.pack('d', item.time) +
                struct.pack('I', 327679) +
                struct.pack('I', item.value))

    replay = item
    return ''.join([
        struct.pack('i', len(replay.frames)),
        struct.pack('i', 0x83),
        struct.pack('i', replay.is_multi),
        struct.pack('i', replay.is_flagtag),
        struct.pack('I', replay.level_id),
        null_padded(replay.level_name, 12),
        struct.pack('i', 0),
        ''.join([struct.pack('f', frame.position.x)
                 for frame in replay.frames]),
        ''.join([struct.pack('f', frame.position.y)
                 for frame in replay.frames]),
        ''.join([struct.pack('h', frame.left_wheel_position.x)
                 for frame in replay.frames]),
        ''.join([struct.pack('h', frame.left_wheel_position.y)
                 for frame in replay.frames]),
        ''.join([struct.pack('h', frame.right_wheel_position.x)
                 for frame in replay.frames]),
        ''.join([struct.pack('h', frame.right_wheel_position.y)
                 for frame in replay.frames]),
        ''.join([struct.pack('h', frame.head_position.x)
                 for frame in replay.frames]),
        ''.join([struct.pack('h', frame.head_position.y)
                 for frame in replay.frames]),
        ''.join([struct.pack('h', frame.rotation)
                 for frame in replay.frames]),
        ''.join([struct.pack('B', frame.left_wheel_rotation)
                 for frame in replay.frames]),
        ''.join([struct.pack('B', frame.right_wheel_rotation)
                 for frame in replay.frames]),
        ''.join([struct.pack('B',
                             frame._gas_and_turn_state & 0b11111100 |
                             (frame.is_turned_right << 1) | frame.is_gasing)
                 for frame in replay.frames]),
        ''.join([struct.pack('h', frame.spring_sound_effect_volume)
                 for frame in replay.frames]),
        struct.pack('I', len(replay.events)),
        ''.join([pack_replay(event) for event in replay.events]),
        struct.pack('I', END_OF_REPLAY_FILE_MARKER),
    ])



from abc import ABCMeta
import random


class Point(object):
    """
    Represent a single 2D point.

    Attributes:
        x (float): The x-coordinate of the point.
        y (float): The y-coordinate of the point.
    """
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __repr__(self):
        return 'Point(x: %s, y: %s)' % (self.x, self.y)

    def __eq__(self, other_point):
        return self.x == other_point.x and self.y == other_point.y


class Obj(object):
    """
    Represent an Elastomania level object, which can be one of: flower, food,
    killer, start.

    Attributes:
        point (Point): The 2D Point that represents the position of the object.
            type (int): The type of the object, which should be one of:
            Obj.FLOWER, Obj.FOOD, Obj.Killer, Obj.START.
        gravity (int): The gravity of the object, if the object is a food
            object. It should be one of: Obj.GRAVITY_NORMAL, Obj.GRAVITY_UP,
            Obj.GRAVITY_DOWN, Obj.GRAVITY_LEFT, Obj,GRAVITY_RIGHT.
        animation_number (int): The animation number of the object.
    """

    FLOWER = 1
    FOOD = 2
    KILLER = 3
    START = 4

    GRAVITY_NORMAL = 0
    GRAVITY_UP = 1
    GRAVITY_DOWN = 2
    GRAVITY_LEFT = 3
    GRAVITY_RIGHT = 4

    def __init__(self, point, type,
                 gravity=GRAVITY_NORMAL, animation_number=1):
        self.point = point
        self.type = type
        self.gravity = gravity
        self.animation_number = animation_number

    def __repr__(self):
        return (
            'Obj(point: %s, type: %s, gravity: %s, animation_number: %s)' %
            (self.point, self.type, self.gravity, self.animation_number))

    def __eq__(self, other_obj):
        return (self.point == other_obj.point and
                self.type == other_obj.type and
                self.gravity == other_obj.gravity and
                self.animation_number == other_obj.animation_number)


class Picture(object):
    """
    Represents an Elastomania level picture.

    Attributes:
        point (Point): The 2D Point that represents the position of the object.
        picture_name (string): The name of the picture resource to use, without
            .PCX, e.g. 'BARREL'.
        picture_name (string): The name of the texture resource to use, without
            .PCX, e.g. 'STONE1'.
        mask_name (string): The name of the texture resource to use, without
            .PCX, e.g. 'MASKHOR'.
        distance (int): The z-ordering distance of the picture. Should be in
            the range 1-999.
        clipping (int): The clipping of the picture. Should be one of:
            Picture.CLIPPING_U, Picture.CLIPPING_G, Picture.CLIPPING_S.
    """

    CLIPPING_U = 0
    CLIPPING_G = 1
    CLIPPING_S = 2

    def __init__(self, point, picture_name='', texture_name='',
                 mask_name='', distance=500, clipping=CLIPPING_U):
        self.point = point
        self.picture_name = picture_name
        self.texture_name = texture_name
        self.mask_name = mask_name
        self.distance = distance
        self.clipping = clipping

    def __repr__(self):
        return (
            ('Picture(point: %s, picture_name: %s, texture_name: %s,' +
             'mask_name: %s, distance: %s, clipping: %s)') %
            (self.point, self.picture_name, self.texture_name,
             self.mask_name, self.distance, self.clipping))

    def __eq__(self, other_picture):
        return (self.point == other_picture.point and
                self.picture_name == other_picture.picture_name and
                self.texture_name == other_picture.texture_name and
                self.mask_name == other_picture.mask_name and
                self.distance == other_picture.distance and
                self.clipping == other_picture.clipping)


class Polygon(object):
    """
    Represents an Elastomania level polygon.

    Attributes:
        points (list): A list of Points defining the polygon contour.
        grass (boolean): A boolean deciding wether or the polygon is a grass
            polygon.
    """
    def __init__(self, points, grass=False):
        self.points = points
        self.grass = grass

    def __repr__(self):
        return 'Polygon(points: %s, grass: %s)' % (self.points, self.grass)


class Level(object):
    """
    Represent an Elastomania level.

    Attributes:
        polygons (list): A list of Polygons in the level.
        objects (list): A list of Objects in the level.
        pictures (list): A list of Pictures in the level.
        level_id (int): A unique unsigned 32bit integer level identifier.
        name (string): The name of level, which should be no longer than 50
            characters long.
        lgr (string): The name of the LGR used for this level, which should be
            no longer than 10 characters long.
        ground_texture (string): The name of the ground texture used for this
            level, which should be no longer than 10 characters long.
        sky_texture (string): The name of the sky texture used for this level,
            which should be no longer than 10 characters long.
    """
    def __init__(self):
        self.polygons = []
        self.objects = []
        self.pictures = []
        self.level_id = random.randint(0, (2 ** 32) - 1)
        self.name = 'Unnamed'
        self.lgr = 'DEFAULT'
        self.ground_texture = 'ground'
        self.sky_texture = 'sky'

    def __repr__(self):
        return (('Level(level_id: %s, name: %s, lgr: %s, ' +
                 'ground_texture: %s, sky_texture: %s)') %
                (self.level_id, self.name, self.lgr,
                 self.ground_texture, self.sky_texture))


class Frame(object):
    """
    Represent a single replay frame.

    Attributes:
        position (Point): The position of the kuski in this frame in level
            coordinates.
        left_wheel_position (Point): The position of the bike's left wheel in
            this frame relative to the position of the kuski.
        right_wheel_position (Point): The position of the bike's right wheel in
            this frame relative to the position of the kuski.
        head_position (point): The position of the kuski's head in this frame
            relative to the position of the kuski.
        rotation (int): The rotation of the kuski in 10000ths of a radian.
        left_wheel_rotation (int): The rotation of the bike's left wheel in
            249/2/pi-ths of a radian.
        right_wheel_rotation (int): The rotation of the bike's right wheel in
            249/2/pi-ths of a radian.
        is_gasing (boolean): Whether or not the bike is gasing in this frame.
        is_turned_right (boolean): Whether or not the bike is turned right in
            this frame.
        spring_sound_effect_volume (int): The spring sound effect volume for
            this frame.
    """
    def __init__(self):
        self.position = Point(0, 0)
        self.left_wheel_position = Point(0, 0)
        self.right_wheel_position = Point(0, 0)
        self.head_position = Point(0, 0)
        self.rotation = 0
        self.left_wheel_rotation = 0
        self.right_wheel_rotation = 0
        self.is_gasing = 0
        self.is_turned_right = 0
        self.spring_sound_effect_volume = 0

    def __repr__(self):
        return ('Frame(position: %s, left_wheel_position: %s, ' +
                'right_wheel_position: %s, head_position: %s, ' +
                'rotation: %s, left_wheel_rotation: %s, ' +
                'right_wheel_rotation: %s, is_gasing: %s, ' +
                'is_turned_right: %s, spring_sound_effect_volume: %s)') % (
                    self.position,
                    self.left_wheel_position,
                    self.right_wheel_position,
                    self.head_position,
                    self.rotation,
                    self.left_wheel_rotation,
                    self.right_wheel_rotation,
                    self.is_gasing,
                    self.is_turned_right,
                    self.spring_sound_effect_volume)


class Event(object):
    """
    Abstract base representation of a single replay event.

    Attributes:
        time (float): The time at which the event occurs, given in 625/273ths
            of a second.
    """
    __metaclass__ = ABCMeta

    def __init__(self):
        self.time = 0

    def __repr__(self):
        return 'Event(time: %s)' % self.time_of_event


class ObjectTouchEvent(Event):
    """
    Represent a single replay object touch event.
    """
    def __init__(self):
        self.object_number = 0

    def __repr__(self):
        return 'ObjectTouchEvent(time: %s, object_number: %s)' % (
               self.time, self.object_number)


class TurnEvent(Event):
    """
    Represent a single replay turn event.
    """
    def __repr__(self):
        return 'TurnEvent(time: %s)' % self.time_of_event


class LeftVoltEvent(Event):
    """
    Represent a single replay left volt event.
    """
    def __repr__(self):
        return 'LeftVoltEvent(time: %s)' % self.time_of_event


class RightVoltEvent(Event):
    """
    Represent a single replay right volt event.
    """
    def __repr__(self):
        return 'RightVoltEvent(time: %s)' % self.time_of_event


class GroundTouchAEvent(Event):
    """
    Represent a single replay ground touch A event.
    """
    def __init__(self):
        self.value = 0

    def __repr__(self):
        return 'GroundTouchAEvent(time: %s)' % self.time_of_event


class GroundTouchBEvent(Event):
    """
    Represent a single replay ground touch B event.
    """
    def __init__(self):
        self.value = 0

    def __repr__(self):
        return 'GroundTouchBEvent(time: %s)' % self.time_of_event


class Replay(object):
    """
    Represent an Elastomania replay.

    Attributes:
        is_multi (boolean): Whether or not the replay is a multiplayer replay.
        is_flagtag (boolean): Whether or not the replay is a flagtag replay.
        level_id (int): The unique identifier of the level this replay is from.
        level_name (string): The name of the level this replay is from.
        frames (list): The frames of this replay.
        events (list): The events of this replay.
    """
    def __init__(self):
        self.is_multi = False
        self.is_flagtag = False
        self.level_id = 0
        self.level_name = ''
        self.frames = []
        self.events = []
        # Needed to preserve unknown bits from rec files
        self._gas_and_turn_state = 0

    def get_exact_duration_in_seconds(self):
        """
        Calculates the exact replay duration in seconds.
        """
        return self.events[-1].time * 625/273

    def __repr__(self):
        return (
            'Replay(is_multi: %s, is_flagtag: %s, level_id: %s, ' +
            'level_name: %s, len(frames): %s, len(events): %s)') % (
            self.is_multi, self.is_flagtag, self.level_id, self.level_name,
            len(self.frames), len(self.events))



from setuptools import setup


setup(
    name='elma',
    version='0.4',
    url='http://github.com/sigvef/elma/',
    license='MIT',
    maintainer='Sigve Sebastian Farstad',
    maintainer_email='sigvefarstad@gmail.com',
    description='Elma Python Library.',
    test_suite="tests",
    packages=['elma']
)
